# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for the GitHub Community Standards Plugin"""

import sys
from pathlib import Path

import pluggy

from RepoAuditor import APP_NAME
from RepoAuditor.Plugins.CommunityStandardsPlugin import GetModule


# ----------------------------------------------------------------------
def test_plugin_imports():
    """Test that the plugin can be imported."""
    # Simply verify that the import worked
    assert GetModule is not None


# ----------------------------------------------------------------------
def test_get_module_returns_object():
    """Test that GetModule returns an object."""
    module = GetModule()
    assert module is not None
    assert hasattr(module, "name")
    assert module.name == "CommunityStandards"


# ----------------------------------------------------------------------
def test_plugin_registration():
    """Test that the plugin can be registered with pluggy."""
    # Create a new plugin manager
    pm = pluggy.PluginManager(APP_NAME)

    # Define a simple hook specification
    class MySpec:
        @pluggy.HookspecMarker(APP_NAME)
        def GetModule(self):
            """Hook specification for GetModule"""

    # Add the hook specification to the plugin manager
    pm.add_hookspecs(MySpec)

    # Register the plugin module
    pm.register(sys.modules[GetModule.__module__])

    # Verify registration worked by checking if we can call the hook
    results = pm.hook.GetModule()
    assert len(results) > 0


# ----------------------------------------------------------------------
class TestCommunityStandardsModule:
    """Unit tests for the GitHubCustomizaton module."""

    def test_Construct(self):
        """Test that the module initializes correctly."""
        module = GetModule()

        assert module is not None
        assert module.name == "CommunityStandards"
        # Comment out checks that would fail with minimal implementation
        # assert module.execution_style.name == "Parallel"
        # assert len(module.queries) == 1
        # assert isinstance(module.queries[0], CommunityStandardsQuery)
        # assert module.requires_explicit_include is False

    def test_GetDynamicArgDefinitions(self):
        """Test GetDynamicArgDefinitions method."""
        module = GetModule()
        dynamic_args = module.GetDynamicArgDefinitions()
        # dynamic_args should be empty dict
        assert dynamic_args.keys() == {"url": "", "pat": "", "branch": ""}.keys()

    def test_GenerateInitialData(self):
        """Test GenerateInitialData method."""
        dynamic_args = {
            "url": "https://github.com/gt-sse-center/RepoAuditor",
            "pat": Path(__file__).parent / "dummy_github_pat.txt",
        }
        module = GetModule()
        updated_dynamic_args = module.GenerateInitialData(dynamic_args)
        # Currently no change in dynamic_args
        assert updated_dynamic_args == dynamic_args


# ----------------------------------------------------------------------
if __name__ == "__main__":
    raise Exception("This should not be run as a script")



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for License.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.License import License
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "license": {
                "name": "MIT License",
            },
        },
    }


@pytest.fixture
def requirement():
    return License()


class TestLicense:
    def test_license_missing(self, requirement, query_data):
        """Check if no "license" key in query_data
        We should get a `IncompleteDataResult`.
        """
        query_data["standard"] = {}
        result = requirement.Evaluate(
            query_data,
            requirement_args={},
        )
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_Empty(self, requirement, query_data):
        """Check if no data under key "license",
        in which case the license value is returned as the empty string.
        """
        query_data["standard"]["license"] = None
        result = requirement.Evaluate(
            query_data,
            requirement_args={"value": requirement.default_value},
        )
        assert result.result == EvaluateResult.Error
        assert "'License' must be set to 'MIT License' (it is currently set to '')." in result.context

    def test_Succesful(self, requirement, query_data):
        """Test successful"""
        result = requirement.Evaluate(
            query_data,
            requirement_args={"value": requirement.default_value},
        )
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the GitHubBaseModule object."""

from pathlib import Path
from typing import Any, Optional
from urllib.parse import urlparse

import requests
import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Module import Module


# ----------------------------------------------------------------------
class GitHubBaseModule(Module):
    """Module for common GitHub settings which can be inherited by various GitHub plugins."""

    # ----------------------------------------------------------------------
    # The __init__ method is inherited from the Module class
    # and does not need to be overriden.

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            "url": (
                str,
                typer.Option(
                    ...,
                    help="[REQUIRED] Github URL (e.g. https://github.com/gt-sse-center/RepoAuditor)",
                ),
            ),
            "pat": (
                str,
                typer.Option(
                    None,
                    help="GitHub Personal Access Token (PAT) or path to a local file containing the PAT.",
                ),
            ),
            "branch": (
                str,
                typer.Option(
                    None,
                    help="Branch to evaluate. The default branch will be used if not specified.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def GenerateInitialData(self, dynamic_args: dict[str, Any]) -> Optional[dict[str, Any]]:
        """Generate the initial data to be used in the `dynamic_args`, such as session info, etc."""
        # Read the PAT (if provided) as a filename or a value.
        # This also decouples the PAT reading logic from the GitHubSession class (which may be mocked for testing).
        github_pat = dynamic_args.get("pat")
        if github_pat:
            potential_file = Path(github_pat)
            if potential_file.is_file():
                with potential_file.open("r") as f:
                    github_pat = f.read().strip()
        # Re-assign github_pat so it can be used within the subclassed modules.
        dynamic_args["pat"] = github_pat

        # Create a GitHub API session
        dynamic_args["session"] = _GitHubSession(dynamic_args["url"], dynamic_args.get("pat"))

        return dynamic_args


# ----------------------------------------------------------------------
# |
# |  Private Types
# |
# ----------------------------------------------------------------------
class _GitHubSession(requests.Session):
    """Session used to communicate with GitHub APIs."""

    # ----------------------------------------------------------------------
    def __init__(
        self,
        github_url: str,
        github_pat: Optional[str],
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.headers.update(
            {
                "X-GitHub-Api-Version": "2022-11-28",
                "Accept": "application/vnd.github+json",
            },
        )

        if github_pat:
            self.headers["Authorization"] = f"Bearer {github_pat}"

        github_url = github_url.removesuffix("/")

        url_parts = urlparse(github_url)
        path_parts = url_parts.path.split("/")

        # The URL should be of the form <github_server>/<username>/<repo>
        if len(path_parts) != 3:  # noqa: PLR2004
            msg = f"'{github_url}' is not a valid GitHub repository URL."
            raise ValueError(msg)

        _, username, repo = path_parts

        if url_parts.netloc.lower() in ["github.com", "www.github.com"]:
            api_url = f"https://api.github.com/repos/{username}/{repo}"
            is_enterprise = False
        else:
            api_url = f"{url_parts.scheme}://{url_parts.netloc}/api/v3/repos/{username}/{repo}"
            is_enterprise = True

        self.github_url = github_url
        self.github_pat = github_pat
        self.github_username = username
        self.github_repository = repo
        self.api_url = api_url
        self.is_enterprise = is_enterprise
        self.has_pat = bool(github_pat)

    # ----------------------------------------------------------------------
    def request(
        self,
        method: str,
        url: str,
        *args,
        **kwargs,
    ) -> requests.Response:
        if url and not url.startswith("/"):
            url = f"/{url}"

        return super().request(
            method,
            f"{self.api_url}{url}",
            *args,
            **kwargs,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for Description.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.Description import Description
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "description": "Description of repository",
        },
    }


@pytest.fixture
def requirement():
    return Description()


class TestDescription:
    """Unit tests for the Description requirement."""

    def test_description_missing(self, requirement, query_data):
        """Test if `description` is missing.
        This happens if no `description` key is present in API response.
        """
        query_data["standard"] = {}
        requirement_args = {"allow-empty": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "The repository's description must be not be empty." in result.context

    def test_description_none(self, requirement, query_data):
        """Test if `description` is None."""
        query_data["standard"]["description"] = None
        requirement_args = {"allow-empty": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "The repository's description must be not be empty." in result.context

    def test_NotAllowEmpty(self, requirement, query_data):
        """Test when description is not allowed to be empty"""
        query_data["standard"]["description"] = ""
        requirement_args = {"allow-empty": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "The repository's description must be not be empty." in result.context

    def test_AllowEmpty(self, requirement, query_data):
        """Test when description is allowed to be empty"""
        query_data["standard"]["description"] = ""
        requirement_args = {"allow-empty": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

    def test_SuccessfulGetDynamicArgDefinitions(self, requirement, query_data):
        """Test using GetDynamicArgDefinitions"""
        requirement_args = {}
        for key, value in requirement.GetDynamicArgDefinitions().items():
            requirement_args[key] = value[1].default
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit test for Executor.py"""

import sys
import textwrap
import time

from typing import cast

import pytest

from dbrownell_Common.Types import override
from dbrownell_Common.TestHelpers.StreamTestHelpers import (
    GenerateDoneManagerAndContent,
    InitializeStreamCapabilities,
)

from RepoAuditor.Display import *
from RepoAuditor.ExecuteModules import *
from RepoAuditor.Module import *
from RepoAuditor.Requirement import *


# ----------------------------------------------------------------------
pytest.fixture(InitializeStreamCapabilities(), scope="session", autouse=True)


# ----------------------------------------------------------------------
class MyModule(Module):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        *args,
        no_initial_data: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.no_initial_data = no_initial_data

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        return {}

    # ----------------------------------------------------------------------
    @override
    def GenerateInitialData(self, dynamic_args: dict[str, Any]) -> Optional[dict[str, Any]]:
        assert not dynamic_args

        if self.no_initial_data:
            return None

        return {"module_name": self.name}

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @override
    def _GetData(self) -> Optional[dict[str, Any]]:
        return {"module_name": self.name}


# ----------------------------------------------------------------------
class MyQuery(Query):
    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
        *args,
        **kwargs,
    ) -> Optional[dict[str, Any]]:
        if args or kwargs:
            super().__init__(*args, **kwargs)

        module_data["query_name"] = self.name
        return module_data


# ----------------------------------------------------------------------
class MyRequirement(Requirement):
    # ----------------------------------------------------------------------
    def __init__(self, result: EvaluateResult, *args) -> None:
        super().__init__(*args)

        self.result = result

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        # Introduce a delay so that we can see things happening
        time.sleep(0.1)

        return Requirement.EvaluateImplResult(
            self.result,
            None,
            provide_resolution=True,
            provide_rationale=True,
        )


# ----------------------------------------------------------------------
class TestExecute:
    # ----------------------------------------------------------------------
    @pytest.mark.parametrize("single_threaded", [False, True])
    def test_Successful(self, single_threaded):
        modules: list[Module] = []

        # ----------------------------------------------------------------------
        def GetExecutionStyle(index: int) -> ExecutionStyle:
            return ExecutionStyle.Parallel if index % 2 == 0 else ExecutionStyle.Sequential

        # ----------------------------------------------------------------------

        for module_index in range(5):
            queries: list[Query] = []

            for query_index in range(4):
                requirements: list[Requirement] = []

                for requirement_index in range(5):
                    requirements.append(
                        MyRequirement(
                            EvaluateResult.Success,
                            f"Requirement-{module_index}-{query_index}-{requirement_index}",
                            "",
                            GetExecutionStyle(requirement_index),
                            "",
                            "",
                        ),
                    )

                queries.append(
                    MyQuery(
                        f"Query-{module_index}-{query_index}",
                        GetExecutionStyle(query_index),
                        requirements,
                    ),
                )

            modules.append(
                MyModule(
                    f"Module-{module_index}",
                    "",
                    GetExecutionStyle(module_index),
                    queries,
                ),
            )

        with DoneManager.Create(sys.stdout, "", line_prefix="") as dm:
            all_results = Execute(
                dm,
                [ModuleInfo(module, {}, {}) for module in modules],
                single_threaded=single_threaded,
            )

            assert dm.result == 0

            for results in all_results:
                assert all(result.result == EvaluateResult.Success for result in results)

    # ----------------------------------------------------------------------
    @pytest.mark.parametrize(
        "data",
        [
            (EvaluateResult.Error, -1, False, False),
            (EvaluateResult.Warning, 1, False, False),
            (EvaluateResult.Warning, -1, True, False),
            (EvaluateResult.Warning, 0, False, True),
            (EvaluateResult.DoesNotApply, 0, False, False),
        ],
    )
    def test_NotSuccess(self, data):
        test_result, expected_result, warnings_as_errors, ignore_warnings = data

        modules: list[Module] = []

        # ----------------------------------------------------------------------
        def GetExecutionStyle(index: int) -> ExecutionStyle:
            return ExecutionStyle.Parallel if index % 2 == 0 else ExecutionStyle.Sequential

        # ----------------------------------------------------------------------

        for module_index in range(2):
            queries: list[Query] = []

            for query_index in range(3):
                requirements: list[Requirement] = []

                for requirement_index in range(4):
                    requirements.append(
                        MyRequirement(
                            (test_result if requirement_index % 3 == 0 else EvaluateResult.Success),
                            f"Requirement-{module_index}-{query_index}-{requirement_index}",
                            "",
                            GetExecutionStyle(requirement_index),
                            "",
                            "",
                        ),
                    )

                queries.append(
                    MyQuery(
                        f"Query-{module_index}-{query_index}",
                        GetExecutionStyle(query_index),
                        requirements,
                    ),
                )

            modules.append(
                MyModule(
                    f"Module-{module_index}",
                    "",
                    GetExecutionStyle(module_index),
                    queries,
                ),
            )

        with DoneManager.Create(sys.stdout, "", line_prefix="") as dm:
            all_results = Execute(
                dm,
                [ModuleInfo(module, {}, {}) for module in modules],
                warnings_as_errors_module_names=(
                    set() if not warnings_as_errors else {module.name for module in modules}
                ),
                ignore_warnings_module_names=(
                    set() if not ignore_warnings else {module.name for module in modules}
                ),
            )

            assert dm.result == expected_result

            # Convert the results into a list of EvaluateResult values
            converted_results: list[EvaluateResult] = []

            for results in all_results:
                converted_results += [result.result for result in results]

            assert converted_results == [
                test_result,
                EvaluateResult.Success,
                EvaluateResult.Success,
                test_result,
                test_result,
                EvaluateResult.Success,
                EvaluateResult.Success,
                test_result,
                test_result,
                EvaluateResult.Success,
                EvaluateResult.Success,
                test_result,
                test_result,
                EvaluateResult.Success,
                EvaluateResult.Success,
                test_result,
                test_result,
                EvaluateResult.Success,
                EvaluateResult.Success,
                test_result,
                test_result,
                EvaluateResult.Success,
                EvaluateResult.Success,
                test_result,
            ]

    # ----------------------------------------------------------------------
    def test_NoModules(self):
        dm_and_content = GenerateDoneManagerAndContent()

        dm = cast(DoneManager, next(dm_and_content))

        all_results = Execute(dm, [])

        assert dm.result > 0, dm.result
        assert all_results == []

        assert cast(str, next(dm_and_content)) == textwrap.dedent(
            """\
            Heading...
              WARNING: There are no modules to process.
            DONE! (1, <scrubbed duration>)
            """,
        )

    # ----------------------------------------------------------------------
    def test_SingleParallel(self):
        dm_and_content = GenerateDoneManagerAndContent()

        all_results = Execute(
            cast(DoneManager, next(dm_and_content)),
            [ModuleInfo(MyModule("MyModule", "", ExecutionStyle.Parallel, []), {}, {})],
        )

        # No queries, so no results
        assert all_results == [[]]

    # ----------------------------------------------------------------------
    def test_NoInitialData(self):
        dm_and_content = GenerateDoneManagerAndContent()

        all_results = Execute(
            cast(DoneManager, next(dm_and_content)),
            [
                ModuleInfo(
                    MyModule("MyModule", "", ExecutionStyle.Sequential, [], no_initial_data=True),
                    {},
                    {},
                )
            ],
        )

        assert all_results == [[]]

        # Assuming major version is 3
        if sys.version_info[1] < 11:
            expected_content = textwrap.dedent(
                """\
            Heading...
              Processing 1 module...
                Processing 'MyModule' (1 of 1)...


                DONE! (ReturnCode.DOESNOTAPPLY, <scrubbed duration>)
              DONE! (ReturnCode.DOESNOTAPPLY, <scrubbed duration>)
            DONE! (ReturnCode.DOESNOTAPPLY, <scrubbed duration>)
            """,
            )
        else:
            # From version 3.11 onwards, IntEnum.__str__ returns the int value.
            expected_content = textwrap.dedent(
                """\
            Heading...
              Processing 1 module...
                Processing 'MyModule' (1 of 1)...


                DONE! (2, <scrubbed duration>)
              DONE! (2, <scrubbed duration>)
            DONE! (2, <scrubbed duration>)
            """,
            )

        assert cast(str, next(dm_and_content)) == expected_content


# ----------------------------------------------------------------------
class TestDisplayResults:
    # ----------------------------------------------------------------------
    @staticmethod
    @pytest.fixture
    def modules() -> list[MyModule]:
        return [
            MyModule(
                "Module1",
                "Module1 description",
                ExecutionStyle.Parallel,
                [
                    MyQuery(
                        "Query1",
                        ExecutionStyle.Parallel,
                        [
                            MyRequirement(
                                EvaluateResult.Success,
                                "Requirement1A",
                                "This is the description for Requirement1A",
                                ExecutionStyle.Parallel,
                                "resolution",
                                "rationale",
                            ),
                            MyRequirement(
                                EvaluateResult.Success,
                                "Requirement1B",
                                "This is the description for Requirement1B",
                                ExecutionStyle.Parallel,
                                "resolution",
                                "rationale",
                            ),
                        ],
                    ),
                ],
            ),
            MyModule(
                "Module2",
                "Module2 description",
                ExecutionStyle.Parallel,
                [
                    MyQuery(
                        "Query2",
                        ExecutionStyle.Parallel,
                        [
                            MyRequirement(
                                EvaluateResult.Success,
                                "Requirement2A",
                                "This is the description for Requirement2A",
                                ExecutionStyle.Parallel,
                                "resolution",
                                "rationale",
                            ),
                        ],
                    ),
                ],
            ),
        ]

    # ----------------------------------------------------------------------
    @pytest.mark.parametrize("display_rationale", [True, False])
    @pytest.mark.parametrize("display_resolution", [True, False])
    @pytest.mark.parametrize("verbose", [False, True])
    @pytest.mark.parametrize(
        "result",
        [
            EvaluateResult.Success,
            EvaluateResult.Warning,
            EvaluateResult.Error,
            EvaluateResult.DoesNotApply,
        ],
    )
    def test_Output(self, result, verbose, display_resolution, display_rationale, modules, capsys, snapshot):
        dm_and_content = GenerateDoneManagerAndContent(verbose=verbose)

        if result in [EvaluateResult.Warning, EvaluateResult.Error]:
            resolution = "resolution"
            rationale = "rationale"
        else:
            resolution = None
            rationale = None

        DisplayResults(
            cast(DoneManager, next(dm_and_content)),
            [
                [
                    Module.EvaluateInfo(
                        result,
                        "Context 1",
                        resolution,
                        rationale,
                        modules[0].queries[0].requirements[0],
                        modules[0].queries[0],
                        modules[0],
                    ),
                    Module.EvaluateInfo(
                        result,
                        "Context 2",
                        resolution,
                        rationale,
                        modules[0].queries[0].requirements[1],
                        modules[0].queries[0],
                        modules[0],
                    ),
                ],
                [
                    Module.EvaluateInfo(
                        result,
                        "Context 3",
                        resolution,
                        rationale,
                        modules[1].queries[0].requirements[0],
                        modules[1].queries[0],
                        modules[1],
                    ),
                ],
            ],
            display_resolution=display_resolution,
            display_rationale=display_rationale,
        )

        content = capsys.readouterr().out

        assert content == snapshot



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the ClassicValueRequirementImpl object."""

import textwrap
from collections.abc import Callable
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.Impl.ValueRequirementImpl import DoesNotApplyResult, ValueRequirementImpl


# ----------------------------------------------------------------------
class ClassicValueRequirementImpl(ValueRequirementImpl):
    """Implementation of a requirement that checks a classic value in GitHub settings."""

    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        default_value: str,
        github_settings_section: Optional[str],
        github_settings_value: Optional[str],
        get_configuration_value_func: Callable[[dict[str, Any]], str | DoesNotApplyResult | None],
        rationale: str,
        subject: Optional[str] = None,
        *,
        requires_explicit_include: bool = False,
    ) -> None:
        if github_settings_section is None:
            resolution = "No resolution instructions are available"
        else:
            if github_settings_value is None:
                github_settings_value_display = "the entity"
            else:
                github_settings_value_display = f"'{github_settings_value}'"

            resolution = textwrap.dedent(
                f"""\
                1) Visit '{{session.github_url}}/settings/branches'
                2) Locate the 'Branch protection rules' section
                3) Click the 'Edit' button next to the branch '{{branch}}'
                4) Locate the '{github_settings_section}' section
                5) Ensure that {github_settings_value_display} is {{__expected_value}}
                """,
            )

        super().__init__(
            name,
            default_value,
            github_settings_value,
            get_configuration_value_func,
            resolution,
            rationale,
            subject,
            requires_explicit_include=requires_explicit_include,
            missing_value_is_warning=False,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireApprovalMostRecentPush object."""

import textwrap
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireApprovalMostRecentPush(ClassicEnableRequirementImpl):
    """Require approval of the most recent reviewable push."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireApprovalMostRecentPush",
            True,
            "disabled",
            "Protect matching branches",
            "Require a pull request before merging -> Require approval of the most recent reviewable push",
            _GetValue,
            textwrap.dedent(
                """\
                The default behavior is to require approval of the most recent reviewable push by someone other than the author of the pull request.

                Reasons for this Default
                ------------------------
                - Self-approval of a pull request eliminates the value provided by a second pair of eyes during a code review.

                Reasons to Override this Default
                --------------------------------
                - You are the only person working on the repository.
                """,
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(
    data: dict[str, Any],
) -> Optional[bool]:
    settings = data["branch_protection_data"].get("required_pull_request_reviews", None)
    if settings is None:
        return None

    return settings.get("require_last_push_approval", None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for MergeCommitMessage.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.MergeCommitMessage import MergeCommitMessage
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "allow_merge_commit": True,
            "merge_commit_message": "BLANK",
        },
    }


@pytest.fixture
def requirement():
    return MergeCommitMessage()


class TestMergeCommitMessage:
    def test_allow_merge_commit_missing(self, requirement, query_data):
        """Test if `allow_merge_commit` is missing"""
        query_data["standard"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_allow_merge_commit_none(self, requirement, query_data):
        """Test if `allow_merge_commit` is None"""
        query_data["standard"]["allow_merge_commit"] = None
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_Disabled(self, requirement, query_data):
        """Test when merge commits are disabled"""
        query_data["standard"]["allow_merge_commit"] = False
        requirement_args = {"value": "BLANK"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert "Merge commits are not enabled" in result.context

    def test_MissingMergeCommitMessage(self, requirement, query_data):
        """Missing merge_commit_message value"""
        query_data["standard"].pop("merge_commit_message", None)
        requirement_args = {"value": "BLANK"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_Successful(self, requirement, query_data):
        """Test successful with BLANK commit message value"""
        requirement_args = {"value": "BLANK"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

    def test_IncorrectCommitMessage(self, requirement, query_data):
        """With incorrect commit message value"""
        query_data["standard"]["merge_commit_message"] = "BLUR"
        requirement_args = {"value": "BLANK"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "it is currently set to 'BLUR'" in result.context



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for SquashMergeCommitMessage.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.SquashMergeCommitMessage import (
    SquashMergeCommitMessage,
)
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "allow_squash_merge": True,
            "squash_merge_commit_message": "COMMIT_MESSAGES",
        },
    }


@pytest.fixture
def requirement():
    return SquashMergeCommitMessage()


class TestSquashMergeCommitMessage:
    def test_allow_merge_commit_missing(self, requirement, query_data):
        """Test if `allow_merge_commit` is missing"""
        query_data["standard"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_Disabled(self, requirement, query_data):
        """Test when squash merges are disabled"""
        query_data["standard"]["allow_squash_merge"] = False
        requirement_args = {"value": "COMMIT_MESSAGES"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert "Squash merge commits are not enabled." in result.context

    def test_MissingValue(self, requirement, query_data):
        """Missing squash_merge_commit_message value"""
        query_data["standard"]["allow_squash_merge"] = True
        query_data["standard"].pop("squash_merge_commit_message", None)

        requirement_args = {"value": "COMMIT_MESSAGES"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_CommitMessage(self, requirement, query_data):
        """With COMMIT_MESSAGES squash-merge commit message value"""
        query_data["standard"]["squash_merge_commit_message"] = "COMMIT_MESSAGES"
        requirement_args = {"value": "COMMIT_MESSAGES"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

    def test_IncorrectValue(self, requirement, query_data):
        """With incorrect commit message value"""
        query_data["standard"]["squash_merge_commit_message"] = "NO_COMMIT_MESSAGES"
        requirement_args = {"value": "COMMIT_MESSAGES"}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "it is currently set to 'NO_COMMIT_MESSAGES'" in result.context



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the ValueRequirementImpl object."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Optional

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.Impl.Common import CreateIncompleteDataResult
from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class DoesNotApplyResult:
    """Evaluate result returned when a requirement does not apply."""

    reason: str


# ----------------------------------------------------------------------
class ValueRequirementImpl(Requirement):
    """Object that implements checking of metadata/settings are set to the specified value."""

    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        default_value: str,
        github_value: Optional[str],
        get_configuration_value_func: Callable[[dict[str, Any]], str | DoesNotApplyResult | None],
        resolution: str,
        rationale: str,
        subject: Optional[str] = None,
        *,
        requires_explicit_include: bool = False,
        missing_value_is_warning: bool = True,
    ) -> None:
        github_value = f"'{github_value}'" if github_value else "the entity"

        if subject is None:
            subject = github_value

        super().__init__(
            name,
            f"Validates that {subject} is set to the expected value.",
            ExecutionStyle.Parallel,
            resolution,
            rationale,
            requires_explicit_include=requires_explicit_include,
        )

        self.github_value = github_value
        self.default_value = default_value
        self.get_configuration_value_func = get_configuration_value_func
        self.missing_value_is_warning = missing_value_is_warning

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            "value": (
                str,
                typer.Option(
                    self.default_value,
                    help=f"Ensures that {self.github_value} is set to the provided value.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        result = self.get_configuration_value_func(query_data)
        if result is None:
            if self.missing_value_is_warning:
                return CreateIncompleteDataResult()

            return Requirement.EvaluateImplResult(EvaluateResult.DoesNotApply, None)

        expected_value = requirement_args["value"]
        is_default_expected_value = expected_value == self.default_value

        if isinstance(result, DoesNotApplyResult):
            if is_default_expected_value:
                return Requirement.EvaluateImplResult(EvaluateResult.DoesNotApply, result.reason)

            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                f"{self.github_value} cannot be set to '{expected_value}' because {result.reason}",
            )

        result = str(result)

        if result != expected_value:
            query_data["__expected_value"] = f"'{expected_value}'"

            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                f"{self.github_value} must be set to '{expected_value}' (it is currently set to '{result}').",
                provide_resolution=True,
                provide_rationale=is_default_expected_value,
            )

        return Requirement.EvaluateImplResult(EvaluateResult.Success, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireUpToDateBranches object."""

import textwrap
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireUpToDateBranches(ClassicEnableRequirementImpl):
    """Require branches to be up to date before merging."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireUpToDateBranches",
            True,
            "disabled",
            "Protect matching branches",
            "Require status checks to pass before merging -> Require branches to be up to date before merging",
            _GetValue,
            textwrap.dedent(
                """\
                The default behavior is to require branches to be up to date before merging. The terminology
                used by GitHub is a bit confusing, as this setting ensures that all GitHub workflows triggered
                by the pull request have completed successfully before the pull request can be merged.

                Reasons for this Default
                ------------------------
                - Workflows run should pass before a pull request is merged into the mainline branch.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(
    data: dict[str, Any],
) -> Optional[bool]:
    settings = data["branch_protection_data"].get("required_status_checks", None)
    if settings is None:
        return None

    return settings.get("strict", None)
